import time
import redis

def redis_init(max_retries=5, retry_delay=1):
    pool = redis.ConnectionPool(host='192.168.100.133', port=6379, db=0)    #ip地址和端口号
    retries = 0
    while retries < max_retries:
        try:
            rds = redis.Redis(connection_pool=pool)
            if rds.ping():
                print("Connected to Redis successfully.")
                return rds
            else:
                print("Redis 连接失败，正在尝试重新连接...")
        except redis.ConnectionError as e:
            print(f"Connection error (retry {retries + 1} of {max_retries}): {e}")
        except Exception as e:
            print(f"An error occurred (retry {retries + 1} of {max_retries}): {e}")

        # 增加延迟时间，防止过于频繁的重试
        time.sleep(retry_delay * (1 ** retries))
        retries += 1

    print("多次尝试重连失败,检查redis服务是否开启.")
    return None

class Publisher:
    def __init__(self):
        self.rds = redis_init()
        if self.rds is None:
            raise Exception("Redis 初始化失败")

    def pub_plan_result(self, result_dict):
        """
        发布路径规划结果
        """
        if self.rds is not None:
            self.rds.publish('ros_plan_result', result_dict)
        else:
            self.redis_init()
            self.rds.publish('ros_plan_result', result_dict)

if __name__ == '__main__':
    Publisher()

